/*----------------------------------------------------------------------
  File    : rbf.h
  Contents: radial basis function network management
  Author  : Christian Borgelt
  History : 2004.08.11 file created
            2013.08.19 adapted to definition of type DIMID in matrix.h
            2018.10.16 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#ifndef __RBF__
#define __RBF__
#include <stdio.h>
#include <math.h>
#include "matrix.h"
#ifdef RBF_PARSE
#ifndef CLS_PARSE
#define CLS_PARSE
#endif
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#endif
#include "cluster.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- training methods --- */
#define RBF_STANDARD    0       /* standard backpropagation */
#define RBF_MANHATTAN   1       /* Manhattan training */
#define RBF_ADAPTIVE    2       /* self-adaptive learning rate */
#define RBF_RESILIENT   3       /* resilient backpropagation */
#define RBF_QUICK       4       /* quick backpropagation */

/* --- description modes --- */
#define RBF_TITLE     0x0001    /* print a title (as a comment) */
#define RBF_INFO      0x0002    /* print add. info. (as a comment) */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- radial basis function network */
  CLSET  *clset;                /* underlying cluster set */
  DIMID  inpcnt;                /* number of inputs */
  DIMID  hidcnt;                /* number of hidden units/clusters */
  DIMID  outcnt;                /* number of outputs */
  DIMID  wgtcnt;                /* number of weights */
  int    method;                /* weight update method */
  double lrate;                 /* learning rate for weights */
  double moment;                /* momentum term */
  double growth;                /* growth    factor */
  double shrink;                /* shrinkage factor */
  double minchg;                /* minimal change (resilient update) */
  double maxchg;                /* maximal change (resilient update) */
  double decay;                 /* weight decay factor */
  MATRIX *mat;                  /* matrix for initialization */
  double *bpes;                 /* backpropagated errors (hidden n.) */
  double *wgts;                 /* connection weights to output layer */
  double *chgs;                 /* connection weight changes */
  double *grds;                 /* (aggregated) weight gradients */
  double *bufs;                 /* connection weight buffers */
  double *outs;                 /* (computed) output values */
  double *trgs;                 /* target values */
  double *errs;                 /* output errors */
  double hids[1];               /* output values of hidden layer */
} RBFNET;                       /* (radial basis function network) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern RBFNET* rbf_create  (CLSET *clset, DIMID outcnt);
extern void    rbf_delete  (RBFNET *rbf, int delcs);
extern CLSET*  rbf_clset   (RBFNET *rbf);
extern DIMID   rbf_inpcnt  (RBFNET *rbf);
extern DIMID   rbf_hidcnt  (RBFNET *rbf);
extern DIMID   rbf_outcnt  (RBFNET *rbf);
extern DIMID   rbf_unitcnt (RBFNET *rbf);
extern DIMID   rbf_wgtcnt  (RBFNET *rbf);

extern NSTATS* rbf_nstats  (RBFNET *rbf);
extern void    rbf_reg     (RBFNET *rbf, const double *ins,
                            const double *trgs, double weight);
extern void    rbf_input   (RBFNET *rbf, DIMID index, double value);
extern void    rbf_target  (RBFNET *rbf, DIMID index, double value);
extern double  rbf_output  (const RBFNET *rbf, DIMID index);

extern void    rbf_clear   (RBFNET *rbf);
extern int     rbf_init    (RBFNET *rbf);
extern void    rbf_method  (RBFNET *rbf, int method);
extern void    rbf_lrate   (RBFNET *rbf, double lrate);
extern void    rbf_moment  (RBFNET *rbf, double moment);
extern void    rbf_factors (RBFNET *rbf, double growth, double shrink);
extern void    rbf_limits  (RBFNET *rbf, double minchg, double maxchg);
extern void    rbf_decay   (RBFNET *rbf, double decay);

extern void    rbf_setup   (RBFNET *rbf);
extern double  rbf_sqrdist (RBFNET *rbf, DIMID hid, const double *vec);
extern double  rbf_dist    (RBFNET *rbf, DIMID hid, const double *vec);
extern void    rbf_exec    (RBFNET *rbf, const double *ins,
                            double *outs);
extern double  rbf_error   (RBFNET *rbf, const double *trgs);
extern double  rbf_bkprop  (RBFNET *rbf, const double *trgs);
extern void    rbf_update  (RBFNET *rbf);

#if 0
#ifdef RBF_TRAIN
extern double  rbf_train   (RBFNET *rbf, MATRIX *matrix, DIMID epochs,
                            DIMID update, double term, DIMID verbose,
                            RANDFN randfn);
#endif
#endif

extern int     rbf_desc    (RBFNET *rbf, FILE *file,
                            int mode, int maxlen);
#ifdef RBF_PARSE
extern RBFNET* rbf_parse   (SCANNER *scan, CLSET *clset,
                            DIMID outcnt, int hdr);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define rbf_clset(n)       ((n)->clset)
#define rbf_inpcnt(n)      ((n)->inpcnt)
#define rbf_hidcnt(n)      ((n)->hidcnt)
#define rbf_outcnt(n)      ((n)->outcnt)
#define rbf_unitcnt(n)     ((n)->inpcnt +(n)->hidcnt +(n)->outcnt)
#define rbf_wgtcnt(n)      ((n)->wgtcnt +((n)->inpcnt+1) *(n)->hidcnt)
#define rbf_nstats(n)      cls_nstats((n)->clset)

#define rbf_sqrdist(n,i,v) cls_sqrdist((n)->clset, i, v)
#define rbf_dist(n,i,v)    cls_dist   ((n)->clset, i, v)

#define rbf_method(n,m)    ((n)->method = (m))
#define rbf_lrate(n,r)     ((n)->lrate  = (r))
#define rbf_moment(n,m)    ((n)->moment = (m))
#define rbf_factors(n,g,s) ((n)->growth = (g), (n)->shrink = (s))
#define rbf_limits(n,a,b)  ((n)->minchg = (a), (n)->maxchg = (b))
#define rbf_decay(n,d)     ((n)->decay  = 1.0 -(d))

#define rbf_input(n,i,v)   cls_value((n)->clset, i, v)
#define rbf_target(n,i,v)  ((n)->trgs[i] = (v))
#define rbf_output(n,i)    ((n)->outs[i])

#endif
